<?php
namespace CGB\Datatables\Controller;

/***************************************************************
 *
 *  Copyright notice
 *
 *  (c) 2017 Christoph Balogh <cb@lustige-informatik.at>
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

/**
 * DatatableController
 */
class DatatableController extends \TYPO3\CMS\Extbase\Mvc\Controller\ActionController
{

    /**
     * dataService
     *
     * @var \CGB\Datatables\Service\DataService
     * @inject
     */
    protected $dataService = NULL;
    
    /**
     * initializeAction
     */
    protected function initializeAction() {
        // TODO: At the moment getting the plugin settings is a wild mixture this should be cleaned up
        $this->contentObj = $this->configurationManager->getContentObject();

        // if no ff values are set try to get from TS
        if ($this->settings['datatables']['listView'] == '') {
            $this->settings['datatables']['listView'] = $this->settings['listView'];
            $this->settings['datatables']['domainObject'] = $this->settings['domainObject'];
            $this->settings['datatables']['config'] = $this->settings['config'];
            $this->settings['datatables']['usergroups'] = $this->settings['usergroups'];
        } 
        
        $this->dataService->setSettings($this->settings, $this->contentObj->data['uid']);
        $settings = $this->dataService->getSettings();
        
        // convert domain model name to table and translation prefix
        $domainModelNameArray = \TYPO3\CMS\Core\Utility\GeneralUtility::trimExplode('\\', $settings['datatables']['domainObject']);
        array_shift($domainModelNameArray);

        // sets the domain object name
        $this->domainModelName = \end($domainModelNameArray);

        // sets the repository name
        $this->dataService->setRepositoryName(str_replace('\\Model\\', '\\Repository\\', $settings['datatables']['domainObject']) . 'Repository');
        
        // set the extension key and the name and the model name. This will be needed in the List-Template
        $extKey = \TYPO3\CMS\Core\Utility\GeneralUtility::camelCaseToLowerCaseUnderscored($domainModelNameArray[0]);
        $this->name = \TYPO3\CMS\Core\Utility\GeneralUtility::camelCaseToLowerCaseUnderscored($this->domainModelName);
        $lowerCaseUnderlined = array_map('\\TYPO3\\CMS\\Core\\Utility\\GeneralUtility::camelCaseToLowerCaseUnderscored', $domainModelNameArray);
        $prefix = 'tx_' . implode('_', $lowerCaseUnderlined);
        
        $this->dataService->setTranslationPrefix($prefix);
        $this->dataService->setTranslationExtKey($extKey);
        $this->dataService->setController($this->domainModelName);
        $this->dataService->setCollapseProperties($settings['datatables']['collapseProperties']);
    }
    
    /**
     * action list
     * @return void
     */
    public function listAction()
    {
        $outputConfig = $this->dataService->renderHeaderList();
        if ($this->dataService->getCollapseProperties()) {
            $collapsePropertyList = array_unique(\TYPO3\CMS\Core\Utility\GeneralUtility::trimExplode(',', $this->dataService->getCollapseProperties()));
            $outputConfig['headers'] = $this->dataService->stack($outputConfig['headers'], $collapsePropertyList);
        }
        
        $settings = $this->dataService->getSettings();
        $neededUsergroups = explode(',', $settings['datatables']['usergroups']);
        $foundUsergroups = $GLOBALS['TSFE']->fe_user->groupData['uid'];
        
        $loggedin = $settings['datatables']['usergroups'] == '';
        $loggedin |= (array_search(-2, $neededUsergroups) !== false) && (count($foundUsergroups) > 0);
        $loggedin |= count(array_intersect($neededUsergroups, $foundUsergroups));
        
        if ($this->dataService->getRequestParams('pageType')) {
            if ($loggedin) {
                return $this->dataService->renderTableData( $outputConfig, $this->name, $this->controllerContext );
            } else {
                return json_encode(array());
            }
        }

        $this->view->assignMultiple(array(
            'extKey' => $this->dataService->getTranslationExtKey(),
            'baseurl' => $GLOBALS['TSFE']->tmpl->setup['config.']['baseURL'],
            'loggedin' => $loggedin,
            'headerList' => $outputConfig['headers'],
            'cObjUid' => $this->dataService->getRequestParams('cObjUid'),
            'settings' => $this->settings,
            'title' => \TYPO3\CMS\Extbase\Utility\LocalizationUtility::translate($this->dataService->getTranslationPrefix() . '.' .  $this->name . '_listing', $this->dataService->getTranslationExtKey()),
            'name' => $this->name,
            'controller' => $this->domainModelName,
        ));
    }
    
}